function [out1,out2,out3] = bruss2d(t,y,flag,n,alpha,InitialValue)
%
% function [out1,out2,out3] = bruss2d(t,y,flag,n,alpha,InitialValue)
%
% The Brusselator example in 2D.
%   t: time.
%   y: space vector.
%   flag: if omitted bruss2d returns the right hand side of the 
%         equation below evaluated at (t,y). If it is the string 'jacobian' 
%         BRUSS2D returns the jacobian at (t,y) otherwise 'init' initializes
%         the problem (it sets the default initial conditions etc.).
%   n: sets the dimension of the problem.
%   alpha: optional, controls the stiffness of the problem (default 1/50).
%   InitialValue: is either the string 'peaks' or 'planes'.
%
% This is a modified form of the Brusselator equation found in:
% E. Hairer and G. Wanner, Solving Ordinary Differential Equations II,
% Stiff and Differential-Algebraic Problems, Springer-Verlag, Berlin,
% 1991, pp. 5-8.
%
% Example: Solve the equation
% 
%   du                               / d^2u   d^2u \
%   -- = A + u^2*v - (B+1)*u + alpha*| ---- + ---- |
%   dt                               \ dx^2   dy^2 /
% 
%   dv                       / d^2v   d^2v \
%   -- = B*u - u^2*v + alpha*| ---- + ---- |
%   dt                       \ dx^2   dy^2 /
%
% with 0 <= x,y <= 1, A = 1, B = 3.4, alpha = 0.02 (default) and
% Neumann boundary conditions and initial conditions
%   u(x,y,0) = peaks(x,y), v(x,y,0) = 0
% for 0 <= t <= 5. 
% Compute the solution on the time grid 
% t = 5*(exp(linspace(0,1,50).^2)-1)/(exp(1)-1).
%
% The spatial derivatives are replaced by finite difference on a grid
% with N^2 mesh points, this gives for N = 30 a system of ODEs of
% dimension 2*N^2 = 1800.
% (The computation takes about 28 seconds an Sun Sparc 20 612MP)
% [b1,b2,b3] = bruss2d([],[],'init',30);
% [t,y] = exp4('bruss2d',b1,b2,b3);
% 
% To see a movie of the solution
% M = plotbruss2d(y,t,1:length(t));
%
% See also HEAT2D, LASER, ROB1D, EXP4, EXP4EX.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

global BRUSS_I1 BRUSS_I2 N2 BRUSS_JAC

if nargin < 3 | isempty(flag)
  % return f(y)
  y = y(:);
  dy = zeros(size(y));
  dy(BRUSS_I1) = -(BRUSS_JAC*y(BRUSS_I1));
  dy(BRUSS_I2) = -(BRUSS_JAC*y(BRUSS_I2));
  tmp = (y(BRUSS_I1).^2).*y(BRUSS_I2);
  dy(BRUSS_I1) = 1 + dy(BRUSS_I1) + tmp - 4.4*y(BRUSS_I1);
  dy(BRUSS_I2) = dy(BRUSS_I2) + 3.4*y(BRUSS_I1) - tmp;
  out1 = dy(:);
else
  switch flag
    case 'init'                       % Return default [tspan,y0,options].
      if ~isempty(n) & isnumeric(n) & (n > 0)
	if nargin < 5
	  alpha = 2e-2;
	end
	if nargin < 6
	  InitialValue='peaks';
	end
	
	out1 = 5*(exp(linspace(0,1,50).^2)-1)/(exp(1)-1);
	N = n;
	N2 = N*N;
	BRUSS_I1 = 1:N2; BRUSS_I2 = N2+1:2*N2;
	fprintf('Setting initial conditions for Brusselator 2D ')
	fprintf('(N = %g and alpha = %g).\n',N,alpha);
	switch InitialValue
	  case 'peaks'
	    out2(BRUSS_I1) = reshape(peaks(N),N2,1);
	    out2(BRUSS_I2) = zeros(N2,1);
	    fprintf('Initial conditions: u = peaks(x,y), v = 0\n');
	  case 'planes'
	    [X,Y] = meshgrid(linspace(0,1,N),linspace(0,1,N));
	    out2(BRUSS_I1) = 0.5 + Y(:);
	    out2(BRUSS_I2) = 1 + 5*X(:);
	    fprintf('Initial conditions: u = 0.5 + y, v = 1 + 5*x\n');
	  otherwise
	    error_msg = sprintf('Unknown Initial conditions requested: %s',...
		InitialValue);
	    error(error_msg);
	end
	out2 = out2(:);
	out3 = odeset('Jacobian','on','stats','on');
	T = spdiags([-ones(N,1),2*ones(N,1),-ones(N,1)],-1:1,N,N);
	T(1,2) = -2;  T(N,N-1) = -2;
	c = alpha * (N-1)^2;
	BRUSS_JAC = c*(kron(T,speye(N,N)) + kron(speye(N,N),T));
      else
	fprintf('The parameter has to be the dimension of the problem!\n');
	error('No initial data set!');
      end
    case 'jacobian'
      % return the jacobian
      tmp = 2*y(BRUSS_I1).*y(BRUSS_I2);
      J11 = -BRUSS_JAC + spdiags(tmp - 4.4,0,N2,N2);
      J12 = spdiags(y(BRUSS_I1).^2,0,N2,N2);
      J22 = -BRUSS_JAC -J12;
      J21 = spdiags(3.4 - tmp,0,N2,N2);
      out1 = [J11 J12; J21 J22];
    otherwise
      error(['Unknown flag ''' flag '''.']);
  end
end
