function [tout,yout] = exp4(odefile,tspan,y,options,varargin)
%
% function [tout,yout] = exp4(odefile,tspan,y,options,varargin)
%
% EXP4:   Solve large systems of (mildly) stiff ordinary differential
%         equations in autonomous form.
%
% SYNTAX: [T,Y] = EXP4('F',TSPAN,Y0,OPTS) with TSPAN = [T0 TFINAL]
%         integrates the autonomous system of differential equations
%			    y' = F(y), y(0) = Y0
%         from time T0 to TFINAL with initial conditions Y0. The
%         function F(y) must return a column vector.  Each row in
%         the solution matrix Y corresponds to a time returned in column
%         vector T.  To obtain solutions at the specific times T0,
%         T1, ..., TFINAL (all increasing), use TSPAN = [T0 T1 ... TFINAL].
%         Y0 is the column vector containing the initial condition.
%         The argument OPTS is a struct of options for the
%         integrator. EXP4 supports the most important features of
%         the matlab ODE solver (see help ODESET). Due to the
%         Krylov techniques used here in this code there are a few
%         other options available for EXP4:
%           * Krylov1, Krylov2, Krylov3: possible values are
%             'phi_arn': Arnoldi process (default)
%             'phi_lan': Lanczos process
%             'phi_lan_sym': Symmetric Lanczos process.
%             They control the Krylov method used in the three
%             stages.
%           * KrylovMax1, KrylovMax2, KrylovMax3: the max. number
%             of Krylov steps allowed in the three stages (the default
%             is 36).
%           * Use2Norm: nonzero forces the use of a weighted
%             2-norm otherwise a weighted inf-norm (default) is used.
%         In order to set these additional options use for example
%         OPTS.Krylov1 = 'phi_lan';
%         to use the Lanczos process in the first stage.
%
% CODE:   This code is an exponential method of order 4 with variable
%         step size selection to solve equations of the form y' =
%         F(y), y0 = y(t0). A dense output formula computes the
%         solution at specified grid points "for free". The
%         computation of phi(A)*v = ((exp(A)-I)/A)*v, where A is the
%         Jacobian matrix of F and v is a vector, is done by Krylov
%         techniques in the code EXP4 and exactly in the code EXP4EX.
%         Therefore EXP4 is the integrator for large systems of
%         ordinary differential equations (say, dimension > 100).
%
%         Theory, implementation issues and numerical experiments can be
%         found in: Marlis Hochbruck, Christian Lubich, Hubert Selhofer:
%         "Exponential Integrators for Large Systems of Differential
%         Equations", to appear in SIAM J. Sci. Comp.
%
%         The latest versions of the codes (matlab and C) and the
%         paper are available at
%
%                    http://na.uni-tuebingen.de/
%
% NOTE:   The system has to be AUTONOMOUS, i.e. the time variable t
%         never explicitly appears in F. However, this is NO
%         restriction, for any problem can easily be reformulated by
%         the following procedure:
%         Say, your right hand side function F looks like
%              function out = ydot(t,y)
%                out(1) = something terms in t,y(1),...,y(N);
%                   ...
%                out(N) = something terms in t,y(1),...,y(N);
%         Introduce now the new component:
%                out(N+1) = 1;
%         (i.e.: the last component of the solution equals the time t)
%         and add Y0(N+1) = T0 to your initial value. Replace t by
%         Y(N+1) in F.
%
% USAGE:  See the help for the examples BRUSS2D, HEAT2D, LASER, ROB1D.
%
% See also Contents, PHIM, PHI_ARN, PHI_LAN, PHI_LAN_SYM.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

global EXP4_DATA EXP4_STATISTICS

% Parse the arguments, options, etc.
IntegratorName = mfilename;
exp4_initialize;

% --------------------------------------------------------------------------
% The main loop
% --------------------------------------------------------------------------

while (t<tfinal) & (t+h>t)

  % stage 1
  if ~reuse_k
    reuse_k_counter = 0;
    % compute Jacobian if necessary
    if (~reuse_J | fail) & old_J_flag
      odefun('NewJacobian',y);
      old_J_flag = 0;
      H = []; V= []; % clear krylov space if the jacobian changes
    end
    % Compute f(y) if necessary
    if (EXP4_STATISTICS.nSteps & ~reject), fy = odefun('YDot',y); end

    [k(:,1:3),num1,fail,V,H,h] = ...
	feval(Krylov1,fy,h,Gamma4,3,KrylovMax1,errw,V,H,1,max(k_guess));
    if ~reject & k_guess_hist > 0
      if k_guess_hist > 1
	k_guess(1:k_guess_hist-1) = k_guess(2:k_guess_hist);
      end
      k_guess(k_guess_hist) = num1;
    end
    num1 = min(size(H));

  else % reuse_k
    reuse_k_counter = reuse_k_counter + 1;
    vprint('======================== reuse k\n');
    % Compute f(y) if necessary, save old value
    if (EXP4_STATISTICS.nSteps & ~reject)
      temp = fy; fy = odefun('YDot',y);
    end

    k_old = k(:,1:3);
    [k(:,1:3),num1,fail] = ...
	feval(Krylov1,fy-temp,h,Gamma4,3,KrylovMax1,errw,[],[],0,0);
    k(:,1:3) = k(:,1:3) + k_old;
  end % stage 1

  if ~reject
    EXP4_STATISTICS.nKrylov1 = [EXP4_STATISTICS.nKrylov1(1)+num1, ...
	  max(EXP4_STATISTICS.nKrylov1(2),num1)];
  end

  % stage 2
  if ~fail
    temp = k(:,1:3) * (h*alpha4)';
    % d4 = fu4                   - fy - A*temp;
    temp = odefun('YDot',y+temp) - fy - odefun('Jacobian*v',temp);

    [k(:,4:6),num2,fail] = ...
	feval(Krylov2,temp,h,Gamma4,3,KrylovMax2,errw,[],[],0,0);

    EXP4_STATISTICS.nKrylov2 = [EXP4_STATISTICS.nKrylov2(1)+num2, ...
	  max(EXP4_STATISTICS.nKrylov2(2),num2)];

    if fail & switch2
      Krylov2 = Krylov2_save;
      KrylovMax2 = KrylovMax2_save;
      vprint('------- 2nd stage, switch back to %s, max steps %3.0f\n',...
	     Krylov2,KrylovMax2);
      switch2 = 0;

      [k(:,4:6),num2,fail] = ...
	  feval(Krylov2,temp,h,Gamma4,3,KrylovMax2,errw,[],[],0,0);

      EXP4_STATISTICS.nKrylov2 = [EXP4_STATISTICS.nKrylov2(1)+num2, ...
	    max(EXP4_STATISTICS.nKrylov2(2),num2)];
    end

    % stage 3
    if ~fail
      num = num2;
      temp = k(:,1:3)*(h*alpha7(1:3))' + (2*h/3)*(k(:,4)+k(:,5)+k(:,6));
      % d7 = fu7                   - fy - A*temp;
      temp = odefun('YDot',y+temp) - fy - odefun('Jacobian*v',temp);

      [k(:,7),num3,fail] = ...
	  feval(Krylov3,temp,h,Gamma4,1,KrylovMax3,errw,[],[],0,0);

      EXP4_STATISTICS.nKrylov3 = [EXP4_STATISTICS.nKrylov3(1)+num3, ...
	    max(EXP4_STATISTICS.nKrylov3(2),num3)];

      if fail & switch3
	Krylov3 = Krylov3_save;
	KrylovMax3 = KrylovMax3_save;
	vprint('------- 3rd stage, switch back to %s, max steps %3.0f\n',...
	      Krylov3,KrylovMax3);
	switch3 = 0;
	[k(:,7),num3,fail] = ...
	    feval(Krylov3,temp,h,Gamma4,1,KrylovMax3,errw,[],[],0,0);

	EXP4_STATISTICS.nKrylov3 = [EXP4_STATISTICS.nKrylov3(1)+num3, ...
	      max(EXP4_STATISTICS.nKrylov3(2),num3)];
      end

      if ~fail
	% compute new approximation at t+h
	temp = k(:,3) + k(:,4) - (4/3)*k(:,5) + k(:,6) + (1/6)*k(:,7);
	y1 = y + h*temp;
	% set the scale for the weighted norm
	sc_norm('set',y,y1);

	% compute error estimate
	% error of embedded W-method
	errw = h*sc_norm('norm',-k(:,1) - k(:,4) + k(:,7) + 2*k(:,2)-temp);
	% error of embedded order 3 method
	err3 = h*sc_norm('norm', ...
			 k(:,3)+(-k(:,4)+k(:,6)+k(:,7))/2-(2/3)*k(:,5)-temp);
	% one is enough
	err = max(eps,min(errw,err3));

	% compute h_err via Gustafsson approach
	h_err = errstep(err, err_acc, h, h_acc, order);

	% compute h_kry
	h_kry = krystep([num1,num2,num3],KrylovHistory)*h;

	% the new step stize
	h_new = min([h_err,h_kry,hmax]);

	vprint(['t: %7.2e, h=%8.2e, errw=%9.2e, err3=%9.2e '...
	      'Krylov: %2.0f,%2.0f,%2.0f (%c)'],...
	    t,h,errw,err3,num1,num2,num3,'K'+(h_kry>=h_err)*('E'-'K'));

	if err<=1
	  % --- accept the error ---
	  old_J_flag = 1;
	  h_acc = h;
	  EXP4_STATISTICS.nSteps = EXP4_STATISTICS.nSteps+1;
	  err_acc = max(1e-2,err);

	  KrylovHistory.num3 = KrylovHistory.num2;
	  KrylovHistory.num2 = KrylovHistory.num1;
	  if reuse_k
	    KrylovHistory.num1(2:3) = [num2 num3];
	  else
	    KrylovHistory.num1 = [num1 num2 num3];
	  end

	  if ~strcmp(Krylov2,'phi_arn') & ...
		max([KrylovHistory.num1(2),KrylovHistory.num2(2),...
		  KrylovHistory.num3(2)]) <= switch2_thres
	    % use Arnoldi in 2nd  stage
	    Krylov2_save = Krylov2;
	    KrylovMax2_save = KrylovMax2;
	    KrylovMax2 = switch2_thres;
	    Krylov2 = 'phi_arn';
	    switch2 = 1;
	    vprint('\n------- 2nd stage, switch to %s, max steps %3.0f',...
		  Krylov2,KrylovMax2);
	  end

	  if ~strcmp(Krylov3,'phi_arn') & ...
		max( [KrylovHistory.num1(3),KrylovHistory.num2(3),...
		  KrylovHistory.num3(3)]) <= switch3_thres
	    % use Arnoldi in 3rd stage
	    Krylov3_save = Krylov3;
	    KrylovMax3_save = KrylovMax3;
	    KrylovMax3 = switch3_thres;
	    Krylov3 = 'phi_arn';
	    switch3 = 1;
	    vprint('\n------- 3rd stage, switch to %s, max steps %3.0f',...
		  Krylov3,KrylovMax3);
	  end

	  V = []; H = [];

	  % write output
          if nargout>0
	    switch EXP4_DATA.Outflag
	      case 1 % output at tspan points
		indx = (t<tspan) & (tspan<t+h);
		indxlen = length(tspan(indx));
		if indxlen
		  s = (tspan(indx) - t)'/h;
		  yout(indx,:) = (y(:,ones(1,indxlen)) + ...
				  k*(h*(dense*cumprod(s(ones(1,4),:))))).';
		  tout(indx) = tspan(indx);
		  nout = nout + indxlen;
		end
		if (nout<length(tspan)) & (tspan(nout+1) == t+h)
		  nout = nout + 1;
		  tout(nout) = t + h;
		  yout(nout,:) = y1.';
		end
	      case 2 % output at every computed points
		nout = nout + 1;
		tout(nout) = t + h;
		yout(nout,:) = y1.';
	      otherwise % output at integration intervall end
		if t + h == tfinal
		  nout = nout+1;
		  tout(nout) = t + h;
		  yout(nout,:) = y1.';
		end
	    end
	  end                           % write output

	  t = t+h; y = y1;

	  % t is already updated here!
	  h_new = min(h_new,abs(tfinal-t));

	  % special treatment if the last step was rejected
	  if reject, reject = 0; h_new = min(h,h_new); end

	  h_old = h; h = h_new;

	else
	  % --- step is rejected ---
	  h_new = min(h_new,abs(tfinal-t));

	  reject = 1;
	  EXP4_STATISTICS.nReject = EXP4_STATISTICS.nReject+1;
	  h_old = h;
	  h = min(h,h_new);
	end

	reuse_J = EXP4_DATA.JConstant ...
	    | ((h_kry <= h_err) & (errw<=1.0) & EXP4_DATA.JIsMatrix ...
	    & (KrylovHistory.num1(1) >= 2*max(num2,num3)));

      end                               % stage 3 did not fail
    end                                 % stage 2 did not fail
  end                                   % stage 1 did not fail

  if fail
    EXP4_STATISTICS.nFails = EXP4_STATISTICS.nFails + 1;
    if reuse_k, V = []; H = []; end
    reject = 1;
    h_old = h;
    h = facmin_kry*h;
  end

  reuse_k = (reuse_J) & (h==h_old) & (~reject);
  if reuse_k_counter > 0
    reuse_k = reuse_k & (num1<KrylovHistory.num1(1)) ...
	      & (reuse_k_counter<=max_reuse_k);
  end
  
  if fail, vprint(' step failed');
  elseif reject
    vprint(' step rejected');
  end
  vprint('\n');

end                                     % main loop (while)

vprint('time: %9.2e\n ',t);

if t < tfinal
  fprintf('\n*** Singularity likely at time t = %e ***\n',t);
end

if nargout > 0
  tout = tout(1:nout);
  yout = yout(1:nout,:);
end

EXP4_STATISTICS.Flops = flops-EXP4_STATISTICS.Flops;
EXP4_STATISTICS.Time = cputime-EXP4_STATISTICS.Time;

fprintf('\nnumber of successful steps: \t%10.5g\n',EXP4_STATISTICS.nSteps);
fprintf('number of rejected steps: \t%10.5g\n',EXP4_STATISTICS.nReject);
fprintf('number of fnc evaluations: \t%10.5g\n',EXP4_STATISTICS.nFuncEvals);
fprintf('number of Jacobi evaluations: \t%10.5g\n',EXP4_STATISTICS.nNewJac);
fprintf('number of Jacobi mult.: \t%10.5g\n',EXP4_STATISTICS.nJacv);
fprintf('number of Jacobi tr. mult.: \t%10.5g\n',EXP4_STATISTICS.nJacTv);
fprintf('number of Krylov steps: \t%10.5g\n',EXP4_STATISTICS.nKrylov1(1)+...
	EXP4_STATISTICS.nKrylov2(1)+EXP4_STATISTICS.nKrylov3(1));
fprintf('number of Krylov failures: \t%10.5g\n',EXP4_STATISTICS.nFails);
fprintf('maximum number of Krylov steps: \t%2.0f, %2.0f, %2.0f\n',...
	EXP4_STATISTICS.nKrylov1(2),EXP4_STATISTICS.nKrylov2(2),...
	EXP4_STATISTICS.nKrylov3(2));
fprintf('number of flops: \t\t%10.5g\n',EXP4_STATISTICS.Flops);
fprintf('%g seconds\n',EXP4_STATISTICS.Time);
fprintf('%g <= AbsTol <= %g, RelTol: %g\n',EXP4_STATISTICS.AtolAndRtol(1),...
	EXP4_STATISTICS.AtolAndRtol(2),EXP4_STATISTICS.AtolAndRtol(3));
