function [tout,yout] = exp4ex(odefile,tspan,y,options,varargin)
%
% function [tout,yout] = exp4ex(odefile,tspan,y,options,varargin)
%
% EXP4EX: Solve systems of (stiff ordinary differential equations
%         in autonomous form.
%
% SYNTAX: [T,Y] = EXP4EX('F',TSPAN,Y0,OPTS) with TSPAN = [T0 TFINAL]
%         integrates the autonomous system of differential equations 
%			    y' = F(y), y(0) = Y0
%         from time T0 to TFINAL with initial conditions Y0. The
%         function F(y) must return a column vector.  Each row in
%         the solution matrix Y corresponds to a time returned in column 
%         vector T.  To obtain solutions at the specific times T0, 
%         T1, ..., TFINAL (all increasing), use TSPAN = [T0 T1 ... TFINAL].
%         Y0 is the column vector containing the initial condition.
%         The argument OPTS is a struct of options for the
%         integrator. EXP4EX supports the most important features of
%         the matlab ODE solver (see help ODESET). 
%
% CODE:   This code is an exponential method of order 4 with variable 
%         step size selection to solve equations of the form y' =
%         F(y), y0 = y(t0). A dense output formula computes the
%         solution at specified grid points "for free". The
%         computation of phi(A)*v = ((exp(A)-I)/A)*v, where A is the
%         Jacobian matrix of F and v is a vector, is done exactly
%         (see help PHIM).
%         Therefore EXP4EX is not suited for large systems of
%         ordinary differential equations (say, dimension > 100).
%         Use it on smaller examples where EXP4 is not well suited.
%
%         Theory, implementation issues and numerical experiments can be
%         found in: Marlis Hochbruck, Christian Lubich, Hubert Selhofer: 
%         "Exponential Integrators for Large Systems of Differential 
%         Equations", to appear in SIAM J. Sci. Comp.
%
%         The latest versions of the codes (matlab and C) and the
%         paper are available at 
%
%                    http://na.uni-tuebingen.de/
%
% NOTE:   The system has to be AUTONOMOUS, i.e. the time variable t 
%         never explicitly appears in F. However, this is NO
%         restriction, for any problem can easily be reformulated by
%         the following procedure:
%         Say, your right hand side function F looks like
%              function out = ydot(t,y)
%                out(1) = something terms in t,y(1),...,y(N);
%                   ...
%                out(N) = something terms in t,y(1),...,y(N);
%         Introduce now the new component:
%                out(N+1) = 1;
%         (i.e.: the last component of the solution equals the time t)
%         and add Y0(N+1) = T0 to your initial value. Replace t by 
%         Y(N+1) in F.
%
% USAGE:  See the help for the examples BRUSS2D, HEAT2D, LASER, ROB1D.
%
% See also Contents, PHIM.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

global EXP4_DATA EXP4_STATISTICS

% Parse the arguments, options, etc.
IntegratorName = mfilename;
exp4_initialize;

notanalyticJ = 0;
% --------------------------------------------------------------------------  
% The main loop
% --------------------------------------------------------------------------  

while (t<tfinal) & (t+h>t)

  if (t + h > tfinal), h = tfinal - t; end
   
  % Compute the slopes
  fy = odefun('YDot',y);
  % Do we need a new Jacobian?
  if ~reuse_J  & old_J_flag & ...
	(~EXP4_DATA.JConstant | EXP4_STATISTICS.nSteps == 0)
    old_J_flag = 0;
    odefun('NewJacobian',y,fy);
    phiA = phim(odefun('ReadJacobian'),h/3);
  else
    % If h changes then we need a new phiA
    if (h_old~=h)
      phiA = phim(odefun('ReadJacobian'),h/3);
    end
  end

  h_old = h;

  k(:,1) = phiA*fy;
  k(:,2) = (h/6)*(phiA*odefun('Jacobian*v',k(:,1))) + k(:,1);     
  k(:,3) = (2*h/9)*(phiA*odefun('Jacobian*v',k(:,2))) + (2/3)*k(:,2) ...
      + (1/3)*k(:,1);
  
  temp = k(:,1:3) * (h*alpha4)';
  % d4 = fu4                   - fy - A*temp;
  temp = odefun('YDot',y+temp) - fy - odefun('Jacobian*v',temp);

  k(:,4) = phiA*temp;
  k(:,5) = (h/6)*(phiA*odefun('Jacobian*v',k(:,4))) + k(:,4);
  k(:,6) = (2*h/9)*(phiA*odefun('Jacobian*v',k(:,5))) + (2/3)*k(:,5) ...
      + (1/3)*k(:,4);
  
  temp = k(:,1:3)*(h*alpha7(1:3))' + (2*h/3)*(k(:,4)+k(:,5)+k(:,6));
  % d7 = fu7                   - fy - A*temp;
  temp = odefun('YDot',y+temp) - fy - odefun('Jacobian*v',temp);

  k(:,7) = phiA*temp;

  temp = k(:,3) + k(:,4) - (4/3)*k(:,5) + k(:,6) + (1/6)*k(:,7);
  y1 = y + h*temp;
  % set the scale for the weighted norm
  sc_norm('set',y,y1);

  % compute error estimate
  % error of embedded W-method
  errw = h*sc_norm('norm',-k(:,1) - k(:,4) + k(:,7) + 2*k(:,2)-temp);
  % error of embedded order 3 method
  err3 = h*sc_norm('norm',k(:,3)+(-k(:,4)+k(:,6)+k(:,7))/2-(2/3)*k(:,5)-temp);
  % one is enough
  err = max(eps,min(errw,err3));

  % compute h_err via Gustafsson approach
  h_err = min(hmax,errstep(err, err_acc, h, h_acc, order));
  
  reuse_J = ((h_err/h_old <= 2) & (err3 < 2*errw) & (errw <= 1.0))...
      | EXP4_DATA.JConstant;
  h_new = h_err;
  
  vprint('t: %7.2e, h=%8.2e, errw=%9.2e, err3=%9.2e ',t,h,errw,err3);


  if err <= 1
    % --- accept the error ---

    EXP4_STATISTICS.nSteps = EXP4_STATISTICS.nSteps+1;
    err_acc = max(1e-2,err); h_acc = h;
    old_J_flag = 1;

    % write output
    if nargout>0
      switch EXP4_DATA.Outflag
	case 1 % output at tspan points
	  indx = (t<tspan) & (tspan<t+h);
	  indxlen = length(tspan(indx));
	  if indxlen
	    s = (tspan(indx) - t)'/h;
	    yout(indx,:) = (y(:,ones(1,indxlen)) + ...
		k*(h*(dense*cumprod(s(ones(1,4),:))))).';
	    tout(indx) = tspan(indx);
	    nout = nout + indxlen;
	  end
	  if (nout<length(tspan)) & (tspan(nout+1) == t+h)
	    nout = nout + 1;
	    tout(nout) = t + h;
	    yout(nout,:) = y1.';
	  end
	case 2 % output at every computed points
	  nout = nout + 1;
	  tout(nout) = t + h;
	  yout(nout,:) = y1.';
	otherwise % output at integration intervall end
	  if t + h == tfinal
	    nout = nout + 1;
	    tout(nout) = t + h;
	    yout(nout,:) = y1.';
	  end
      end
    end                                 % write output
    
    t = t + h; y = y1;

    % t is already updated here!
    h_new = min(h_new,abs(tfinal-t));

    % special treatment if the last step was rejected
    if reject, reject = 0; h_new = min(h,h_new); end
	  
    vprint('\n');

    h = h_new;
  else
    % --- step is rejected ---
    vprint(' step rejected.\n');
    EXP4_STATISTICS.nReject = EXP4_STATISTICS.nReject + 1;
    reject = 1;
    h = min(h,h_new);
  end

end                                     % while (main loop)

vprint('t: %7.2e\n ',t);

if t < tfinal
  fprintf('\n*** Singularity likely at time t = %e ***\n',t);
end

if nargout > 0
  tout = tout(1:nout); 
  yout = yout(1:nout,:);
end

EXP4_STATISTICS.Flops = flops-EXP4_STATISTICS.Flops;
EXP4_STATISTICS.Time = cputime-EXP4_STATISTICS.Time;

fprintf('\nnumber of successful steps: \t%10.5g\n',EXP4_STATISTICS.nSteps);
fprintf('number of rejected steps: \t%10.5g\n',EXP4_STATISTICS.nReject);
fprintf('number of fnc evaluations: \t%10.5g\n',EXP4_STATISTICS.nFuncEvals);
fprintf('number of Jacobi evaluations: \t%10.5g\n',EXP4_STATISTICS.nNewJac);
fprintf('number of flops: \t\t%10.5g\n',EXP4_STATISTICS.Flops)
fprintf('%g seconds\n',EXP4_STATISTICS.Time);
fprintf('%g <= AbsTol <= %g, RelTol: %g\n',EXP4_STATISTICS.AtolAndRtol(1),...
    EXP4_STATISTICS.AtolAndRtol(2),EXP4_STATISTICS.AtolAndRtol(3));
