function [h,f0] = h_init(tspan,y0,ord)

% function [h,f0] = h_init(tspan,y0,ord)
%
% This function computes (guesses) a starting step size for a Runge
% Kutta type method of order ord.
%    tspan: is the interval of integration, 
%    y0: is the initial data, i.e. y(tspan(0)) = y0,
%    ord: is the order of the Runge Kutta type method.
% 
% This strategy costs two function evaluations, but f0 = f(t0,y0)
% can be reused in the algorithm.
%
% See Hairer, Norsett, Wanner: "Solving Ordinary Differential 
% Equations I. Nonstiff Problems", Springer 1992. (page 169ff)
%
% See also EXP4, EXP4EX.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

global EXP4_DATA  
  
%
% A first guess
%
f0 = odefun('YDot',y0);
d0 = sc_norm('norm',y0);
d1 = sc_norm('norm',f0);
if (d0 > 1e-5) & (d1 > 1e-5)
  h0 = 0.01*(d0/d1);
else
  h0 = 1e-6;
end
%
% One Euler-step
%
y1 = y0 + h0*f0;
%
% A second guess
% 
f1 = odefun('YDot',y1);
d2 = sc_norm('norm',f1-f0,y1)/h0;
md = max(d1,d2);
if md<1e-15
  h1 = max(1e-6,h0*1e-3);
else
  h1 = 1/(100*md)^(1/(ord+1));
end

h = min(100*h0,h1);
