function [out1,out2,out3] = heat2d(t,y,flag,n,alpha)
%
% function [out1,out2,out3] = heat2d(t,y,flag,n,alpha)
%
% The heat equation in 2D.
%   t: time.
%   y: space vector.
%   flag: if omitted heat2d returns the right hand side of the 
%         equation below evaluated at (t,y). If it is the string 'jacobian' 
%         heat2d returns the jacobian at (t,y) otherwise 'init' initializes
%         the problem (it sets the default initial conditions etc.).
%   n: sets the dimension of the problem.
%   alpha: optional, controls the stiffness of the problem (default 1/50).
%
% Example: Solve the equation
% 
%   du         / d^2u   d^2u \
%   -- = alpha*| ---- + ---- |
%   dt         \ dx^2   dy^2 /
%
% with 0 <= x,y <= 1, alpha = 0.02 (default) and
% Neumann boundary conditions and initial conditions
%   u(x,y,0) = peaks(x,y),
% for 0 <= t <= 10.
%
% Compute the solution on the time grid 
% t = linspace(0,10,50)
%
% The spatial derivatives are replaced by finite differences on a grid
% with N^2 mesh points, this gives for N = 30 a system of ODEs of
% dimension N^2 = 900.
% (The computation takes about 11 seconds an Sun Sparc 20 612MP)
% [b1,b2,b3] = heat2d([],[],'init',30);
% [t,y] = exp4('heat2d',linspace(0,10,50),b2,b3);
%
% See also BRUSS2D, LASER, ROB1D, EXP4, EXP4EX.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

global HEAT_JAC

% we have no flag
if nargin < 3 | isempty(flag)
  % return f(y), this is the default
  N2 = length(HEAT_JAC);

  y = y(:);
  dy = zeros(size(y));
  dy = HEAT_JAC*y;                      % remember, it's a linear problem
  
  out1 = dy(:);
  
else
  % we do have a flag
  switch flag 
    case 'init'                         % Return default [tspan,y0,options].
      if ~isempty(n) & isnumeric(n) & (n > 0)
	% is alpha sepcified?
	if nargin < 5
	  alpha = 2e-2;                 % the default
	end
	
	out1 = [0; 10];                 % time interval of integration
	N = n;
	N2 = N^2;
	i1 = 1:N2; 
        fprintf('Setting initial conditions for the heat equation in 2D\n')
        fprintf('(N = %g and alpha = %g).\n',N,alpha);
	out2(i1) = reshape(peaks(N),N2,1); % the initial condition
	fprintf('Initial conditions: u = peaks(x,y), v = 0\n');
	out2 = out2(:);
	out3 = odeset('Jacobian','on','stats','on');
	% Here we compute the jacobian once and for all times
	T = spdiags( [ -ones(N,1),2*ones(N,1),-ones(N,1)],-1:1,N,N);
	T(1,2) = -2;  T(N,N-1) = -2;
	c = alpha * (N-1)^2;
	fprintf('Stiffness parameter alpha = %g\n',alpha);
	HEAT_JAC = (-c)*(kron(T,speye(N,N)) + kron(speye(N,N),T));
      else
	fprintf('The parameter has to be the dimension of the problem!\n');
	error('No initial data set!');
      end
    
    case 'jacobian'
      out1 = HEAT_JAC;                  % return the already computed jac.
      
    otherwise
      error(['Unknown flag ''' flag '''.']);
  end
end
