function [out1,out2,out3] = laser(t,y,flag,N,K)
%
% function [out1,out2,out3] = laser(t,y,flag,N,K)
%
% This is a time-dependent Schr"odinger equation.
% It models an atom/molecule interacting with a high intensity laser.
% See Peskin, Kosloff, Moiseyev: "Polynomial propagators for the
% (t,t') method", J. Chem. Phys., Vol. 100, No. 12, p 8849-8855.
%
% We solve the modified equation for u = u(x,t):
%
%     du
%   i*-- = H(x,t)*u
%     dt
% 
%   with the Hamiltonian
%
%              1 d^2      x^2
%   H(x,t) = - -*---- + K*--- + mu*sin^2(w*t)*x .
%              2 dx^2      2
%
%   where mu = 100, w = 1 and K = 10 (as default values).
%   The initial value is u(x,0) = exp(-sqrt(K)*x^2/2) on the space
%   interval [-10,10]. Semi-discretization in space is done by a 
%   pseudospectral method with N Fourier modes with periodic boundary
%   conditions.
%
% Example: The computation takes about 4 minutes on a Sun
% Sparc 20 612MP.
% N = 512;
% [l1,l2,l3] = laser([],[],'init',N);
% [t,y] = exp4('laser',l1,l2,l3);
% A short "movie" of the solution:
% x = linspace(-10,10-20/N,N);
% for j = 1:length(t)
%   plot(x,real(y(j,1:N)),x,abs(y(j,1:N)),'m');
%   axis([-10 10 -1 1]);
%   S = sprintf('real part and abs. value of the solution at time=%3.2f',t(j));
%   title(S);
%   drawnow;
% end
%
% This example shows the feature of a matrix free computation of the
% product A*v of the Jacobian A and a vector v.
%
% See also BRUSS2D, HEAT2D, ROB1D, EXP4, EXP4EX.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

% i = sqrt(-1);

global LASER_K LASER_OMEGA LASER_MU LASER_X LASER_A LASER_FFTP1 LASER_FFTP2

if nargin < 3 | isempty(flag)
  % Return dy/dt = F(t,y).
  
  % the last entry is the time variable
  y = y(:);
  N = length(y)-1;
  t1 = y(N+1);
  S2 = LASER_MU*sin(LASER_OMEGA*t1)^2;
  
  dy = zeros(size(y));
  dy(1:N) = -ifft(LASER_FFTP1 .*fft(y(1:N)));
  dy(1:N) = -(i/2)*(dy(1:N)+(LASER_FFTP2+(2*S2)*LASER_X).*y(1:N));
  dy(N+1) = 1.0;
  
  out1 = dy(:);
else
  switch flag
    case 'init' % Return default [tspan,y0,options].
      if ~isempty(N) & isnumeric(N) & (N > 0)
	if nargin < 4
	  N = 512;
	end
	
	if nargin < 5
	  LASER_K = 10;                 % default K = 10
	else
	  LASER_K = K;
	end
	
	fprintf('Setting initial conditions for the laser-example, ');
	fprintf('N = %g, K = %5.2f.\n',N,LASER_K);

	LASER_A = 10;                   % space interval, default [-10,10]
	dx = 2*LASER_A/N;
	LASER_OMEGA = 1;
	LASER_MU = 100;

	LASER_X = linspace(-LASER_A,LASER_A-dx,N)';
	LASER_FFTP1 = -([0:N/2-1 -N/2:-1].'*pi/LASER_A).^2;
	LASER_FFTP2 = LASER_K*LASER_X.^2;
 
	out1 = linspace(0,1,25);        % solve for t = 0..1, 25 grid points
	out2 = [ exp(-sqrt(LASER_K)*LASER_X.^2/2); 0 ];
	out2 = out2(:);
	out3 = odeset('Jacobian','off','abstol',1e-6,'reltol',1e-6, ...
	    'stats','on');
	out3.IsComplex = 'on';
      else
	fprintf('The parameter has to be the dimension of the problem!\n');
	error('No initial data set!');
      end
    case 'Jacobian*v'
      v = y(:,2);
      y = y(:,1);
      
      N = length(y)-1;
      
      t1 = y(N+1);
      S2 = LASER_MU*sin(LASER_OMEGA*t1)^2;
      
      out1 = zeros(size(y));
      out1(1:N) = -ifft(LASER_FFTP1.*fft(v(1:N)));
      out1(1:N) = -(i/2)*(out1(1:N)+(LASER_FFTP2+(2*S2)*LASER_X).*v(1:N))...
	  - (i*v(N+1)*LASER_OMEGA*LASER_MU*sin(2*LASER_OMEGA*t1))...
	  *LASER_X.*y(1:N);
      out1(N+1) = 0;
      
    case 'JacobianT*v'
      v = y(:,2);
      y = y(:,1);
      
      N = length(y)-1;
      
      t1 = y(N+1);
      S2 = LASER_MU*sin(LASER_OMEGA*t1)^2;
      
      out1 = zeros(size(y));
      out1(1:N) = -ifft(LASER_FFTP1.*fft(v(1:N)));
      out1(1:N) = (i/2)*( out1(1:N) + (LASER_FFTP2 + (2*S2)*LASER_X).*v(1:N));
      out1(N+1) = (i*LASER_OMEGA*LASER_MU*sin(2*LASER_OMEGA*t1)) ...
	  *((LASER_X.*y(1:N))' *v(1:N));
      
    otherwise
      error(['Unknown flag ''' flag '''.']);
  end
end
