function out = odefun(flag,y,fy)
%
% function out = odefun(flag,y,fy)
%
% This is a utility function for EXP4 and EXP4EX to unify the
% computation of problem dependent terms, i.e. the function
% evaluation, the computation of the jacobian and the computation of
% the jacobian times a vector. It automagically decides how to compute
% these terms: using a user supplied jacobian matrix times vector
% routine, compute the jacobian analytically or numerically (only for
% EXP4EX) and so on. It also updates the statistics (number of
% function evaluations, ...).
%
% Arguments:
%   flag: possible values are
%      'YDot': evaluate the function F(y).
%      'NewJacobian': compute the jacobian.
%      'ReadJacobian': return the jacobian without any computation.
%      'Jacobian*v': return jacobian*y.
%      'JacobianT*v': return jacobian^T*y.
%   y: A vector of the dimension of the ODE system.
%   fy: The right hand side of the ODE evaluated at y, F(y).

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

global EXP4_DATA EXP4_STATISTICS

t = 'please do not use the time variable, exp4 is for autonomous systems';

switch flag
  case 'YDot'
    out = feval(EXP4_DATA.Name,t,y,'',EXP4_DATA.VarArg{:});
    EXP4_STATISTICS.nFuncEvals = EXP4_STATISTICS.nFuncEvals + 1;
    
  case 'NewJacobian'
    if EXP4_DATA.JIsMatrix
      if (~EXP4_DATA.JConstant) | isempty(EXP4_DATA.Jacobian)
	vprint('calculating new Jacobian\n')
	EXP4_DATA.Jacobian = feval(EXP4_DATA.Name,t,y,'jacobian',...
	    EXP4_DATA.VarArg{:});
	EXP4_STATISTICS.nNewJac = EXP4_STATISTICS.nNewJac + 1;
      end
      out = [];
    else
      if nargin==3 % only exp4ex calls odefun with 3 input arguments
	global ODEFUN_FAC ODEFUN_G
	vprint('calculating new Jacobian numerically\n')
	[EXP4_DATA.Jacobian,ODEFUN_FAC,ODEFUN_G,nfeval] = ...
	    numjac(EXP4_DATA.Name,0,y,fy,...
	    EXP4_DATA.Threshold*EXP4_DATA.Rtol,ODEFUN_FAC,[],[],...
	    ODEFUN_G,EXP4_DATA.VarArg{:});
	EXP4_STATISTICS.nFuncEvals = EXP4_STATISTICS.nFuncEvals + nfeval;
	EXP4_STATISTICS.nNewJac = EXP4_STATISTICS.nNewJac + 1;
      else
	EXP4_DATA.Jacobian = [];
	EXP4_DATA.JacobianEvalPoint = y;
	out = [];
      end
    end
    
  case 'ReadJacobian'
    if nargout==1
      out = EXP4_DATA.Jacobian;
    else
      error('One output argument expected.')
    end
    
  case 'Jacobian*v'
    if isempty(EXP4_DATA.Jacobian)
      out = feval(EXP4_DATA.Name,t,[EXP4_DATA.JacobianEvalPoint y], ...
	  'Jacobian*v',EXP4_DATA.VarArg{:});
    else
      out = EXP4_DATA.Jacobian*y;
    end
    EXP4_STATISTICS.nJacv = EXP4_STATISTICS.nJacv + 1;
    
  case 'JacobianT*v'
    if isempty(EXP4_DATA.Jacobian)
      out = feval(EXP4_DATA.Name,t,[EXP4_DATA.JacobianEvalPoint y], ...
	  'JacobianT*v',EXP4_DATA.VarArg{:});
    else
      out = EXP4_DATA.Jacobian'*y;
    end
    EXP4_STATISTICS.nJacTv = EXP4_STATISTICS.nJacTv + 1;
    
  otherwise
    error(['Unknown flag ''' flag '''.']);
end
