function [x,k,fail,V,H,h] = ...
    phi_arn(v,h,Gamma4,fac_t,max_m,errw,V,H,ASSRFlag,k_guess)
%
% function [x,k,fail,V,H,h] = ...
%     phi_arn(v,h,Gamma4,fac_t,max_m,errw,V,H,ASSRFlag,k_guess)
%
% This functions computes the Arnoldi approximation to 
%     phi(t*A)v with phi(z) = (exp(z)-1)/z.
% The matrix A is given implicitly.
% The matrix evaluation of the function phi must be in phim.m
% The convergence criterion is based on a generalized residual.
%
% See also PHI_LAN, PHI_LAN_SYM.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

% fprintf('---------------------------------------------------------------\n');
% fprintf('This is the Arnoldi process to approximate phi(A)v\n');

global EXP4_DATA 

% Initialize
norm_v = norm(v);
tau    = h*Gamma4;
k      = size(V,2);

if (k~=0) & (length(H)~=0) 		% step rejected, reuse Krylov space

  z = phim(H(1:k,1:k),tau,fac_t)*norm_v;
  x = V(:,1:k)*z(1:k,:); 		% x(:,j) approximates \phi(j*t A) 
  
  k = 0;
  fail = 0; 				% no fail
  return
end

if (round(fac_t) ~= fac_t) | (fac_t<1) | (fac_t>3)
  fprintf('\ninvalid parameters in Arnoldi process\n');
  error('Program stopped!');
end

if ~ASSRFlag | norm_v==0
  if tau*fac_t*sc_norm('norm',v) < 1e-3
    x = zeros(EXP4_DATA.Neq,fac_t);
    fail = 0;
    k = 0;
    return
  end
end

z = zeros(max_m,fac_t);
V = zeros(EXP4_DATA.Neq,max_m);
H = zeros(max_m+1,max_m);
res = 0;
V(:,1) = v/norm_v;
k = 1;
finished = 0; fail = 0;

while ~finished & k<=max_m & ~fail

  % tmp = A*V(:,k);
  tmp = odefun('Jacobian*v',V(:,k));

  for j=1:k
    h_nkp1 = V(:,j)'*tmp;
    tmp    = tmp - V(:,j)*h_nkp1;
    H(j,k) = h_nkp1;
  end
  h_nkp1 = norm(tmp);
  H(k+1,k) = h_nkp1;
  
  % compute approximation of phi(A)v: x = ||v|| V*phi(H)e_1
  
  if min(abs(EXP4_DATA.KrylovTestIndex - k)) == 0 | h_nkp1 < 1e-10
%%%%%  if min(abs(EXP4_DATA.KrylovTestIndex - k)) == 0 | (h_nkp1*tau) < 1e-10
    
    z = phim(H(1:k,1:k),tau,fac_t)*norm_v;

    if ~(any(isinf(z(:))) | any(isnan(z(:))))
      scaled_norm = h*sc_norm('norm',tmp);
      res = scaled_norm*tau*fac_t*h*abs(z(k,fac_t));
    
      if k==1
	res = 5*res;
      end
      
      %fprintf('step: %2.0f, res= %9.2e  \n',k,res*max(1,errw));
      
      finished = (res*max(1,errw)<1);
    else
      fprintf('######################## NAN or INF warning');
      fprintf(' # step: %2.0f ##########\n',k);

      if k == max_m & ASSRFlag
	scaled_norm = h*sc_norm('norm',tmp);
	res = scaled_norm*tau*fac_t*h*abs(z(k,fac_t));
      end
      
    end

  end
    
  V(:,k+1) = tmp/h_nkp1;
  k = k+1;
end

k = k-1;
fail = ~finished | fail;

if finished
  H = H(1:k+1,1:k);
  V = V(:,1:k);
  x = V(:,1:k)*z(1:k,:); 		% x(:,j) approximates \phi(j*t A)
else
  if ASSRFlag % allow step size reduction
    [z,h] = kry_step_red(H,h,scaled_norm,norm_v,errw,res,z,Gamma4,fac_t);
    fail = 0;
    V = V(:,1:k);
    x = V(:,1:k)*z(1:k,:); 		% x(:,j) approximates \phi(j*t A) 
  else
    fail = 1;
    x = zeros(EXP4_DATA.Neq,fac_t);
  end
end

