function [x,k,fail,V,H,h] = ...
    phi_lan_sym(v,h,Gamma4,fac_t,max_m,errw,V,H,ASSRFlag,k_guess)
%
% function [x,k,fail,V,H,h] = ...
%       phi_lan_sym(v,h,Gamma4,fac_t,max_m,errw,V,H,ASSRFlag,k_guess)
%
% This functions computes the Lanczos approximation to 
%     phi(t*A)v with phi(z) = (exp(z)-1)/z.
% The matrix A is given implicitly and A = A' is assumed, so
% the Hermitian Lanczos process is used.
% The matrix evaluation of the function phi must be in phim.m.
% The convergence criterion is based on a generalized residual.
%
% See also PHI_ARN, PHI_LAN, LANCZOS2.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

% fprintf('---------------------------------------------------------------\n');
% fprintf('This is the Hermitian Lanczos process to approximate phi(A)v\n');

global EXP4_DATA

% Initialize
norm_v = norm(v);
tau    = h*Gamma4;
k      = size(H,2);

if (k~=0) & (length(H)~=0) 		% step rejected, reuse Krylov space

  z = phim(H(1:k,1:k),tau,fac_t)*norm_v;
  x = lanczos2(v/norm_v,H,z,k);
  
  fail = 0; 				% no fail
  return
end

if (round(fac_t) ~= fac_t) | (fac_t<1) | (fac_t>3)
  fprintf('\ninvalid parameters in Lanczos process\n');
  error('Program stopped!');
end

if ~ASSRFlag | norm_v==0
  if tau*fac_t*sc_norm('norm',v)<1e-3
    x = zeros(EXP4_DATA.Neq,fac_t);
    fail = 0;
    k = 0;
    return
  end
end

z = zeros(max_m,fac_t);
%%V = 1;
H = zeros(max_m+1,max_m);
res = 0;
v = v/norm_v;
V = v;
vk = v;
k = 1;
norm_vk = 1;
vk_vk_old = 1;
finished = 0; fail = 0;

while ~finished & k<=max_m & ~fail
  
  % tmpv = A*vk;
  tmpv = odefun('Jacobian*v',vk);
  vk_A_vk = vk'*tmpv;
  beta = norm_vk;
  alpha =  vk_A_vk;

  if k > 1 
    tmpv = tmpv - beta * vkm1 - alpha* vk;
    H(k-1,k) = beta;
  else
    tmpv = tmpv - alpha* vk;
  end
  H(k,k)   = alpha;
  
  % compute approximation of phi(A)v: x = ||v|| V*phi(H)e_1
  
  if min(abs(EXP4_DATA.KrylovTestIndex - k)) == 0 | h_nkp1 < 1e-10
    
    z = phim(H(1:k,1:k),tau,fac_t)*norm_v;

    if ~(any(isinf(z(:))) | any(isnan(z(:))))
      scaled_norm = h*sc_norm('norm',tmpv);
      res = scaled_norm*tau*fac_t*h*abs(z(k,fac_t));
    
      if k==1
	res = 5*res;
      end
    
      %fprintf('step: %2.0f, res= %9.2e',k,res);
      %fprintf('\n');
    
      finished = (res<1/max(1,errw));
    else
      fprintf('######################## NAN or INF warning');
      fprintf(' # step: %2.0f ##########\n',k);
    
      if k == max_m & ASSRFlag
	scaled_norm = h*sc_norm('norm',tmp);
	res = scaled_norm*tau*fac_t*h*abs(z(k,fac_t));
      end
    end
  end
  
  if ~finished

    % compute norms of Lanczos vectors and (k+1,k) entry of H
    norm_vk  = norm(tmpv);
    H(k+1,k) = norm_vk;
    vkm1 = vk;
    vk   = tmpv/norm_vk;
    k = k+1;
    V(:,k) = vk;
  end 
  
end    

fail = ~finished | fail;

if finished
  H = H(1:k+1,1:k);
  %  x = V(:,1:k)*z(1:k,:); 		% x(:,j) approximates \phi(j*t A)
  x = lanczos2(v,H,z,k);
else
  if ASSRFlag % allow step size redcution
    [z,h] = kry_step_red(H,h,scaled_norm,norm_v,errw,res,z,Gamma4,fac_t);
    fail = 0;
    x = lanczos2(v,H,z,k);
  else
    fail = 1;
    x = zeros(EXP4_DATA.Neq,fac_t);
  end
end
