function out = phim(A,h,fac_t,v)
%
% function out = phim(A,h,fac_t,v)
%
% Matrix evaluation of the function
%     phi(z) = (exp(z)-1)/z
% via Pade approximation, diagonalization, or via an expansion of the
% matrix A and the matlab EXPM function (see internal function
% PHI_EXPAND).
%
% Called with arguments:    Return:
% A                         the matrix phi(A)
% A, h                      the matrix phi(h*A)
% A, h, fac_t               the matrix out(:,j)=phi(j*h*A)*e1, j=1..fac_t
% A, h, fac_t, v            the matrix out(:,j)=phi(j*h*A)*v, j=1..fac_t

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

if nargin <= 2
  if nargin == 1;
    h = 1;
  end
  % compute phi(h*A) and return the full matrix
  out = phim_full(A,h);
else
  % compute phi(j*h*A) for j=1..fac_t<=3
  if nargin == 3
    % compute phi(h*A)*e1
    if fac_t == 1
      out = phie1(A,h);
    else
      [out,Exp] = phie1(A,h);
      out = recurse(fac_t,out,Exp);
    end
  else
    % compute phi(h*A)*v
    if fac_t == 1
      out = phim_full(A,h)*v;
    else
      [Phi,Exp] = phim_full(A,h);
      out = recurse(fac_t,Phi*v,Exp);
    end
  end
end

%
% Internal functions
%

function z = recurse(fac_t,z1,Exp)
%
% compute phi(j*A)*z1 for j=1..fac_t<3.
%
z = zeros(length(z1),fac_t);
z(:,1) = z1;
if fac_t >= 2
  z(:,2) = (Exp*z1 + z1)/2;
  if fac_t >= 3
    z(:,3) = (2/3)*(Exp*z(:,2)) + (1/3)*z1;
  end
end

      
function [Phi,Exp] = phie1(A,h)
%
% Compute phi(h*A)*e1 and expm(h*A).
%
% Strategy: If the dimension of A is smaller than 9 use diagonalization
% or Pade approximation, otherwise expand the matrix and use EXPM.
%
n = size(A,1);
  
if n == 1
  Phi = phis(h*A);
  if nargout > 1
    Exp = exp(h*A);
  end
elseif n <= 8
  % Scale A by power of 2 so that its norm is < 1/2 .
  s = max(0,nextpow2(h*norm(A,'inf'))+1);
  
  if nargout < 2
    if s >= 6
      Phi = phim_diag(A,h,eye(n,1));
      if isempty(Phi)
	Phi = phim_pade(A,h,s);
	Phi = Phi(:,1);
      end
    else
      Phi = phim_pade(A,h,s);
      Phi = Phi(:,1);
    end
  else
    if s >= 6
      [Phi,Exp] = phim_diag(A,h,eye(n,1));
      if isempty(Phi)
	[Phi,Exp] = phim_pade(A,h,s);
	Phi = Phi(:,1);
      end
    else
      [Phi,Exp] = phim_pade(A,h,s);
      Phi = Phi(:,1);
    end
  end
else
  if nargout < 2
    Phi = phi_expand(A,h);
  else
    [Phi,Exp] = phi_expand(A,h);
  end
end
      
      
function [Phi,Exp] = phi_expand(A,h)
%
% Compute phi(h*A) by expanding the matrix and using EXPM.
%
n = size(A,1);
if h == 1
  E = expm([A eye(n,1);zeros(1,n+1)]);
else
  E = expm([h*A eye(n,1);zeros(1,n+1)]);
end
Phi = E(1:n,n+1);
Exp = E(1:n,1:n);
  

function [Phi,Exp] = phim_full(A,h)
%
% Compute phi(h*A) and expm(h*A).
%
% Strategy: Use diagonalization or Pade approximation.
%
if size(A) == [1 1]
  Phi = phis(h*A);
  if nargout > 1
    Exp = exp(h*A);
  end
else
  % Scale A by power of 2 so that its norm is < 1/2 .
  s = max(0,nextpow2(h*norm(A,'inf'))+1);
  
  if nargout < 2
    if s >= 6
      Phi = phim_diag(A,h);
      if isempty(Phi)
	Phi = phim_pade(A,h,s);
      end
    else
      Phi = phim_pade(A,h,s);
    end
  else
    if s >= 6
      [Phi,Exp] = phim_diag(A,h);
      if isempty(Phi)
	[Phi,Exp] = phim_pade(A,h,s);
      end
    else
      [Phi,Exp] = phim_pade(A,h,s);
    end
  end
end


function [Phi,Exp] = phim_pade(A,h,s)
%
% Compute phi(h*A) by (6,6)-Pade approximation. s is the scaling
% factor such that (1/2)^s * ||h*A|| < 1/2.
%
global PHIM_PADE_N PHIM_PADE_D PHIM_PADE_Q

if isempty(PHIM_PADE_N)
  PHIM_PADE_N = [ 1/26  5/156 1/858  1/5720 1/205920 1/8648640 ];
  PHIM_PADE_D = [ -6/13 5/52  -5/429 1/1144 -1/25740 1/1235520 ];
  PHIM_PADE_Q = 6;
end

% the scaling
A = A*(h/(2^s));

X = A;
Id = speye(size(A));
Phi = Id + PHIM_PADE_N(1)*A;
D   = Id + PHIM_PADE_D(1)*A;
for k = 2:PHIM_PADE_Q
  X   = A*X;
  Phi = Phi + PHIM_PADE_N(k)*X;
  D   = D   + PHIM_PADE_D(k)*X;
end
Phi = D\Phi;
D = []; X = [];

% Undo scaling
if s > 0
  Exp = A*Phi + Id;
  Phi = (Exp+Id)*Phi/2;
  for k=2:s
    Exp = Exp*Exp;
    Phi = (Exp+Id)*Phi/2;
  end

  if nargout == 2
    Exp = Exp*Exp;
  end
else
  if nargout == 2
    Exp = A*Phi + Id;
  end
end


function [Phi,Exp] = phim_diag(A,h,v)
%
% Compute phi(h*A)*v and exp(h*A) by diagonalization.
%
n = size(A,1);

% Do diagonalization
[V,D] = eig(full(A));

if cond(full(V))<1e6
  D = h*diag(D);
  if nargin < 3
    if isreal(A);
      Phi = real(V*sparse([1:n],[1:n],phis(D))/V);
    else
      Phi = V*sparse([1:n],[1:n],phis(D))/V;
    end
  else
    if isreal(A) & isreal(v)
      Phi = real(V*(phis(D).*(V\v)));
    else
      Phi = V*(phis(D).*(V\v));
    end
  end
  if nargout == 2
    if isreal(A);
      Exp = real(V*sparse([1:n],[1:n],exp(D))/V);
    else
      Exp = V*sparse([1:n],[1:n],exp(D))/V;
    end
  end
else
  Phi = [];
  if nargout == 2
    Exp = [];
  end
end


function v = phis(v)
%
% Compute phi(v) for a vector v. If |v(i)| < 1/2 use Pade
% approximation otherwise compute it directly.
%
indx = abs(v)<0.5;
  
v(indx) = phis_pade(v(indx));  
indx    = ~indx;
v(indx) = (exp(v(indx))-1)./v(indx);


function v = phis_pade(v)
%
% Compute phi(v) for a vector v using the Pade approximation.
%
global PHIM_PADE_N PHIM_PADE_D PHIM_PADE_Q

if isempty(PHIM_PADE_N)
  PHIM_PADE_N = [ 1/26  5/156 1/858  1/5720 1/205920 1/8648640 ];
  PHIM_PADE_D = [ -6/13 5/52  -5/429 1/1144 -1/25740 1/1235520 ];
  PHIM_PADE_Q = 6;
end

X = v;
Id = ones(size(v));
Phi = Id + PHIM_PADE_N(1)*v;
D   = Id + PHIM_PADE_D(1)*v;
for k = 2:PHIM_PADE_Q
  X   = v.*X;
  Phi = Phi + PHIM_PADE_N(k)*X;
  D   = D   + PHIM_PADE_D(k)*X;
end
v = Phi./D;
