function [out1,out2,out3] = rob1d(t,y,flag,n)
%
% function [out1,out2,out3] = rob1d(t,y,flag,n)
%
% The Robertson equation in 1D.
%   t: time.
%   y: space vector.
%   flag: if omitted rob1d returns the right hand side of the 
%         equation below evaluated at (t,y). If it is the string 'jacobian' 
%         rob1d returns the jacobian at (t,y) otherwise 'init' initializes
%         the problem (it sets the default initial conditions etc.).
%   n: sets the dimension of the problem.
%
% This a modification of the Robertson example of Hindmarsh and Byrne.
% E. Hairer and G. Wanner, Solving Ordinary Differential Equations II,
% Stiff and Differential-Algebraic Problems, Springer-Verlag, Berlin,
% 1991, p3.
%
% Example: Solve the equation
% 
%   du                                       d^2u
%   -- = -0.04*u + 1e4*v*w           + alpha*----
%   dt                                       dx^2
% 
%   dv                                       d^2v
%   -- =  0.04*u - 1e4*v*w - 3e7*v^2 + alpha*----
%   dt                                       dx^2
%
%   dw                                       d^2v
%   -- =                     3e7*v^2 + alpha*----
%   dt                                       dx^2
%
% with 0 <= x,y <= 1, A = 1, B = 3.4, alpha = 1/50 (default) and
% Neumann boundary conditions and initial conditions
%   u(x,0) = 1 + sin(x), v(x,0) = w(x,0) = 0
% for 0 <= t <= 100.
% Compute the solution on the time grid 
% t = linspace(0,100,100)
%
% The spatial derivatives are replaced by finite difference on a grid
% with N^2 mesh points, this gives for N = 30 a system of ODEs of
% dimension 3*N = 90.
% [r1,r2,r3] = rob1d([],[],'init',30);
% [t,y] = exp4ex('rob1d',linspace(0,100,100),r2,r3);
%
% See also BRUSS2D, HEAT2D, LASER, EXP4, EXP4EX.

% This file is part of the exp4 package available at
% http://na.uni-tuebingen.de.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% file COPYING of the distribution or send email to
% exp4@na.uni-tuebingen.de.

global ROB_JAC ROB_I1 ROB_I2 ROB_I3 ROB_N

if nargin < 3 | isempty(flag)
  y = y(:);
  dy = zeros(size(y));
  dy(ROB_I1) = -0.04*y(ROB_I1) + 1e4*y(ROB_I2).*y(ROB_I3);
  dy(ROB_I3) = 3e7* y(ROB_I2).^2;
  dy(ROB_I2) = -dy(ROB_I1) - dy(ROB_I3) + ROB_JAC*y(ROB_I2);
  dy(ROB_I1) = dy(ROB_I1) + ROB_JAC*y(ROB_I1);
  dy(ROB_I3) = dy(ROB_I3) + ROB_JAC*y(ROB_I3);
  out1 = dy(:);
else
  switch flag
    case 'init'                       % Return default [tspan,y0,options].
      if ~isempty(n) & isnumeric(n) & (n > 0)
	ROB_N = n;
	ROB_I1 = 1:ROB_N; ROB_I2 = ROB_N+1:2*ROB_N; ROB_I3 = 2*ROB_N+1:3*ROB_N;
	h = 1/(ROB_N-1);
	alpha = 1/50;
	c = alpha/h^2;

	fprintf('Setting initial conditions for Robertson 1D, N = %g.\n',ROB_N);
	out1 = [0; 0.04e9];
	out2 = [1 + sin(linspace(0,2*pi,ROB_N)) zeros(1,ROB_N) zeros(1,ROB_N)];
	out2 = out2(:);
	out3 = odeset('Jacobian','on','stats','on');
	T = spdiags( [ -ones(ROB_N,1),2*ones(ROB_N,1),-ones(ROB_N,1)],...
	    -1:1,ROB_N,ROB_N);
	% boundary conditions
	T(1,2) = -2; T(ROB_N,ROB_N-1) = -2;
	ROB_JAC = (-c)*T;
      else
	fprintf('The parameter has to be the dimension of the problem!\n');
	error('No initial data set!');
      end
    case 'jacobian'
      J21 = spdiags(0.04 + zeros(ROB_N,1),0,ROB_N,ROB_N);
      J11 = ROB_JAC - J21;
      J12 = spdiags(1e4*y(ROB_I3),0,ROB_N,ROB_N);
      J13 = spdiags(1e4*y(ROB_I2),0,ROB_N,ROB_N);
      J32 = spdiags(6e7*y(ROB_I2),0,ROB_N,ROB_N);
      J22 = ROB_JAC - J12 - J32;
      J23 = -J13;
      J31 = sparse(ROB_N,ROB_N);
      J33 = ROB_JAC;
      out1 = [J11 J12 J13; J21 J22 J23; J31 J32 J33];
    otherwise
      error(['Unknown flag ''' flag '''.']);
  end
end
